#!/usr/bin/env python

__author__	="mforster"
__date__	="$Dec 15, 2011 1:00 PM$"
__version__	= "v1.4.3"

import sys
import time
import resource
import argparse
import os.path

import pysam

from textwrap import dedent


def main():
	""" The main function dealing with the parser and checking arguments.
	    (Is being called at the end of this script.)"""

	setlimits()

	#
	# parser + description
	#
	parser = argparse.ArgumentParser(prog="pibase_to_vcf",
	formatter_class=argparse.RawDescriptionHelpFormatter,
	description=dedent("""
	about:
	  Convert a single-sample pibase file into VCFv4.1 format.
	"""), epilog=dedent("""

	quality scores:
	  The BestQual can have 8 different levels and which are associated
	  with PHRED-like scores as following:

	  Score (in %)   BestQual
	    30   99.9%   no-question-mark-tag, confirmation from both strands
	    20   99%     no-question-mark-tag
	    15	 97%     ?1	
	     7	 80%     ?2, ?3, ?4, ?5 and ?6
	     3	 50%     ?7 and ?8

	  With -u you can set the scores to your own need.

	example:
	  Output BestGen genotype and BestQual quality as a PHRED-like quality score. 
	  User-define the quality-scores for BestQuals of " " with/without
	  both-stranded confirmation as 40 and 20, and "?1",...,"?8" as 15,...,3:

	  pibase_to_vcf -u 40 20 15 10 7 7 7 7 3 3 in.txt out.txt mysample mysample.bam

	authors:
	  Michael Forster (m.forster@ikmb.uni-kiel.de)
	  Ingo Thomsen (i.thomsen@ikmb.uni-kiel.de)

	version:
	  """ + __version__ + """ (23. Feb 2012)""")
	, version=__version__)


	#
	# arguments
	#
	parser.add_argument("infile" , help=dedent("""
		single-sample file from pibase_consensus, pibase_flagsnp, 
		pibase_annot, or pibase_gen_from_snpacts"""))

	parser.add_argument("outfile", help="VCF output file name.")

	parser.add_argument("samplename",  help="name of sample, as it should appear in the VCF file.")

	parser.add_argument("bamfile", help="original bam file (needed for original \
									 contig names, as pibase uses numbers)")


	#
	# options
	#
	parser.add_argument("-d", action="store_true", help="if, set only diploid \
			and haploid calls are written to the VCF file - triploid calls will \
			be filtered out.")

	parser.add_argument("-u", nargs=10, metavar="s", help="define the\
			PHRED-like scores for the 10 possible BestQuals (see below)")
	#
	# parse and check arguments
	#
	args = parser.parse_args()

	if not os.path.exists(args.bamfile):
		sys.exit ("*** Error: BAM file does not exist: " + args.bamfile)

	if not os.path.exists(args.infile):
		sys.exit ("*** Error: pibase input file does not exist: " + args.infile)

	if args.u:
		try:
			scores = list(map(int, args.u))				
		except ValueError:
			sys.exit("*** Error: the user-defined PHRED-likes scores must be integers > 0: " 
					+ " ".join(args.u))
	else:
		scores = [ 30,20,15,7,7,7,7,7,3,3 ]

	# now, start extraction
	pib_tovcf(args.infile, args.outfile, args.samplename, args.bamfile, scores,	args.u, args.d)


###############################################################################
#
# FUNCTIONS
#
###############################################################################

def setlimits():
	#############################################################################
	# sets resource limits (mandatory for job interoperability on cluster)
	# to 10MB Stack + 1GB data + 1GB RLIMIT_RSS
	# first steps taken from
	# http://docs.python.org/library/resource.html and
	# http://www.doughellmann.com/PyMOTW/resource/
	# Author: MICHAEL FORSTER m.forster@ikmb.uni-kiel.de
	# Version: 13 September 2010
	#############################################################################

	#############################################################################
	# Adjust the maximum size (in bytes) of the process's heap
	#############################################################################

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit starts as  :', soft

	# 1GB heap:
	newheap=1024*1024*1024

	resource.setrlimit(resource.RLIMIT_DATA, (newheap, hard))

	soft, hard = resource.getrlimit(resource.RLIMIT_DATA)
	#print 'Data/heap limit changed to :', soft

	#############################################################################
	# ca 10MB original stack changed to exactly 10MB stack:
	#############################################################################

	#newstack=10*1024*1024+1
	newstack=10*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit starts as  :', soft
	resource.setrlimit(resource.RLIMIT_STACK, (newstack, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_STACK)
	#print 'Stack limit changed to :', soft

	#############################################################################
	# 1GB RLIMIT_RSS
	# "Maximale resident set size in Byte. Bei Speicherengpaessen entzieht der Kern
	# den Prozessen, die den resident set size ueberschreiten, den zuviel
	# angeforderten Speicher."
	# http://pronix.linuxdelta.de/C/Linuxprogrammierung/Linuxsystemprogrammieren_C_Kurs_Kapitel3a.shtml
	#############################################################################

	newrss=1024*1024*1024
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS starts as  :', soft
	resource.setrlimit(resource.RLIMIT_RSS, (newrss, hard))
	soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
	#print 'RLIMIT_RSS changed to :', soft

def pib_tosnpactsown (out, chrom, chr, pos1, ref, bestgen):
	###########################################################################
	# generate a line in snpacts 'own' format and write into filehandle out
	# col 1 - chromosome name ( -pchr 1 )
	# col 2 - 1-based position ( -sp 1 -psnp 2 )
	# col 3 - ref allele ( -prb 3 )
	# col 4 - best allele ( -pbb 4 )
	# col 5 - second best allele ( -psb 5 )
	# col 6 - chromosome name (pchr = 1)
	# snpacts-import needs to be invoked (from cmd line) as follows: 
	# snpact -ft own -hg hg18 -all (...above 6 switches...) -f file -t table
	###########################################################################
	line = chrom[chr] + '\t' + str(pos1) + '\t'  + ref + '\t'
	bestallele = ref
	secondallele = ref
	muttype = 'hom'
	if bestgen != '':
		bestallele = bestgen [:1]
		secondallele = bestgen [1:]
		if bestallele != secondallele:
			muttype = 'het'
	line += bestallele + '\t' + secondallele + '\t' + muttype + '\n'
	out.write(line)

def pib_tosnpacts1 (infile, inf, ic, maxcol, n, k):
	###########################################################################
	# called by pib_tosnpacts: to read and process a single line from infile
	###########################################################################

	line = inf.readline()
	# print line
	if len(line) == 0:
		# EOF
		if k == 1:
			return 0,0,'',''
		else:
			# pibase_diff
			return 0,0,'','',''
	cols=line.split('\t')
	ncols = len(cols)
	if ncols < maxcol:
		# end of table
		if k == 1:
			return 0,0,'',''
		else:
			# pibase_diff
			return 0,0,'','',''
	try:
		chr = int (cols[ ic[0] ])
		pos1 = int (cols[ ic[1]  ])
	except:
		print ' *** Error: Unable to read integer values (chrom column and pos column)'
		print ' Line:', line[:30]
		print ' cols[ic[0]], cols[pos_col]: ', cols[ic[0]], cols[ic[1]]
		sys.exit (' Details: chrom-column %s, pos-column %s, file %s, line %s' % (ic[0]+1, ic[1]+1, infile, n))
	ref = cols[ ic[2] ]
	if len(ref) != 1 or 'ACGTNacgtn'.find(ref)<0 :
		print ' *** Error: Bad reference base (not length 1) or not in ACGTNacgtn:'
		print ' Line:', line[:30]
		print ' ref: ', ref
		sys.exit (' file %s, line number %s' % (infile, n))
		
	bestgen = cols[ ic[3] ]		
	if len(bestgen) != 2: 
		if bestgen != '':
			print ' *** Warning: Bad BestGen (not length 2): BestGen = %s' % ( bestgen )
			# print ' Line:', line[:30]
			print ' *** File: %s   Line number: %s' % (infile, n)
			print ' *** Replacing with Reference (only in order to get snpacts locus annotation) and continuing normally.' 
		bestgen = ref+ref
	if k == 1:
		# print 'chr, pos1, ref, bestgen :', chr, pos1, ref, bestgen
		return chr, pos1, ref, bestgen
	else:
		# pibase_diff
		# print 'chr, pos1, ref, bestgen, #D:', chr, pos1, ref, bestgen, cols[0]
		return chr, pos1, ref, bestgen, cols[0]

def pib_tovcf (infile, outfile, samplename, bamfile, scores, userdefined,	diploidonly):
	#  http://www.1000genomes.org/wiki/Analysis/Variant%20Call%20Format/vcf-variant-call-format-version-41

	#######################################################################################
	# Echo command line arguments 
	#######################################################################################
	
	print 'Infile : %s' % (infile) 
	print 'Outfile: %s' % (outfile) 
	print 'BAMfile: %s' % (bamfile)
	print 'Sample name in VCF file: %s' % (samplename)

	# assign the scores to different variables ...
	q0b = scores[0]
	q0  = scores[1]
	qother = {}
	t = []
	for i in range(1,len(scores)-1):
		key="?"+str(i)
		value = scores[1+i]
		t.append(str(value) + "='" + key +"'")
		qother[key]=value

	# ... and print it
	print "Quality output: PHRED-like SNP quality scores:"
	print "%s = ' ' with both-stranded confirmation, %s = ' ' without both-stranded confirmation" % tuple(scores[0:2])
	print ",".join(t)
	if userdefined:
		print "(These are your user-defined values for conversion of pibase BestQual/strand tags to PHRED-like scores)"
	else:
		print "(These are the default values for conversion of pibase BestQual/strand tags to PHRED-like scores)"

	if diploidonly:
		print "Only haploid and diploid calls are shown"
	print 'Quality output: None ("." in VCF file).'

	#######################################################################################
	# open bamfile, extract reference sequence chromosomes (or contigs) and refseq file
	#######################################################################################

	samfile = pysam.Samfile( bamfile, "rb")
	
	# Get names of contigs/chromosomes
	refs = samfile.references
	
	# Get nested-dictionary of BAM header
	d=samfile.header
	
	# Get nested dictionary of SQ-lines
	dd=d['SQ']
	
	# Get reference sequence file name from first SQ-line:
	refname = dd[0]['UR'].split(':')[1]


	############################################################################
	# open infile and determine columns for BestGen, BestQual, chr, pos, ref
	# from the header line
	############################################################################
	
	try:
		inf = open(infile)
	except:
		sys.exit(" *** Error: Could not open pibase input file: " + infile)

	cols = False	
	for line in inf:

		parts = line.split("\t")
		if parts[0].startswith("#"):

			# could be comment or header line
			# assume it's a header and extract the column numbers

			# remove the beginning "#"
			parts[0] = parts[0].lstrip("#").lstrip()
			try:
				cols = {}
				cols["bestgen"]  = parts.index("BestGen")
				cols["bestqual"] = parts.index("BestQual")
				cols["pos"]      = parts.index("Pos[1]")
				cols["ref"]		 = parts.index("Ref")
				cols["a+-"]      = parts.index("A+-")
				cols["b+-"]      = parts.index("B+-")
				try:
					cols["chr"] = parts.index("Chr")
				except ValueError:
					cols["chr"] = parts.index("C")
				break
			except ValueError:
				cols = False
				pass  # ... because it seems not to be a header line

	if cols:
		print "Numbers for columns of interest:", cols
	else:
		sys.exit(" *** Error: Could not find a header line with all columns of interest in " + infile)


	############################################################################
	# write the header of the vcf file
	############################################################################

	try:
		outf = open(outfile, "w")
	except:
		sys.exit(" *** Error: Could not open file for writing: " + outfile)

	# simple local output wrapper
	def p(s): 
		print >> outf, s


	p("##fileformat=VCFv4.1")
	p("##fileDate=" + time.strftime("%Y%m%d"))
	p("##source=pibase_to_vcf_" + __version__)
	p("##input=" + infile)
	p("##reference=" + refname)
	p('##FORMAT=<ID=GT,Number=1,Type=String,Description="Genotype">')
	p('##FORMAT=<ID=GQ,Number=1,Type=Integer,Description="Genotype Quality (PHRED-like)>')
	p('##FORMAT=<ID=GQP,Number=1,Type=Integer,Description="Genotype Quality Pibase: "PA" or "PB" = passed (with both strand confirmation) OR "?x" = failed, with x a integer 1-8>')

	p("#" + "\t".join(("CHROM", "POS", "ID", "REF", "ALT", "QUAL","FILTER","INFO", "FORMAT", samplename)))
	
	############################################################################
	# loop over remaining lines of infile
	############################################################################

	for line in inf:

		parts = line.split("\t")
		out = []

		#
		# CHR
		#
		out.append(parts[cols["chr"]])

		#
		# POS (it's 1-based in vcf and pibase)
		#
		out.append(parts[cols["pos"]])

		#
		# ID (normally, there should be no variant ID in the pibase file)
		#
		out.append(".")

		#
		# REF		
		ref = parts[cols["ref"]]
		out.append(ref)

		#
		# ALT
		#
		allele = list(parts[cols["bestgen"]])
		if len(allele) == 0:
			continue

		# the set of possible alleles (without the reference)
		alt = list(set(allele))
		try:
			alt.remove(ref)
		except ValueError:
			pass

		if len(alt) == 0:
			out.append(".")
		else:
			out.append(",".join(alt))

		#
		# QUAL 
		#
		out.append(".")

		#
		# FILTER
		#
		qual = parts[cols["bestqual"]]
		if qual == "":
			out.append("PASS")
			if parts[cols["a+-"]] == "+-" and parts[cols["b+-"]] == "+-":
				phred = q0b
				gqp="PB"
			else:
				phred = q0
				gqp="PA"
		else:
			out.append("FAIL")
			phred = qother[qual]
			gqp = qual

		#
		# INFO
		#
		out.append(".")

		#
		# FORMAT
		#
		out.append("GT:GQ:GQP")
		format = []

		# the genotype information
		t = []	
		for b in allele:
			if b == ref:
				t.append("0")
			else:
				t.append(str(alt.index(b)+1))
		if diploidonly and len(t) > 2:
			continue
		format.append("/".join(t))

		# the genotype qualities
		format.append(str(phred))
		format.append(gqp)

		# add it to the main output list
		out.append(":". join(format))

		#
		# finally, the output for this line
		#
		p("\t".join(out))


#
# start main, if script is being executed
#
if __name__ == "__main__": main()
